;; Generated package description from yasnippet.el  -*- no-byte-compile: t -*-
(define-package "yasnippet" "0.14.1.0.20240215.153445" "Yet another snippet extension for Emacs" '((cl-lib "0.5") (emacs "24.4")) :commit "5b565c54b0dce77f6f0e4dd12111f43379e26d8d" :maintainer '("Noam Postavsky" . "npostavs@gmail.com") :keywords '("convenience" "emulation") :url "http://github.com/joaotavora/yasnippet")
